package de.ozgcloud.bescheid;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.info.BuildProperties;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.OzgCloudCreateSubCommandsRequest;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.bescheid.attacheditem.AttachedItem;
import de.ozgcloud.bescheid.attacheditem.AttachedItemService;
import de.ozgcloud.bescheid.attributes.ClientAttributeService;
import de.ozgcloud.bescheid.common.user.UserProfileService;
import de.ozgcloud.bescheid.vorgang.Vorgang;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.bescheid.vorgang.VorgangService;
import de.ozgcloud.command.Command;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.Document;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
@RequiredArgsConstructor
class BescheidService {

	private static final String ERROR_MESSAGE_NO_SERVICE = "No Bescheid Endpoint is configured.";

	static final String VORGANG_BESCHEIDEN_ORDER = "VORGANG_BESCHEIDEN";
	static final String SUBCOMMANDS_EXECUTION_MODE = "PARALLEL";
	static final String SEND_POSTFACH_NACHRICHT_ORDER = "SEND_POSTFACH_NACHRICHT";

	static final String FIELD_POSTFACH_ID = "postfachId";
	static final String FIELD_REPLY_OPTION = "replyOption";
	static final String REPLY_OPTION = "FORBIDDEN";
	static final String FIELD_SUBJECT = "subject";
	static final String FIELD_MAIL_BODY = "mailBody";
	static final String FIELD_ATTACHMENTS = "attachments";
	static final String SUBJECT = "Ihr Bescheid zum Antrag";

	private final VorgangService vorgangService;
	private final AttachedItemService attachedItemService;
	private final UserProfileService userProfileService;
	@Qualifier(BescheidManagerConfiguration.COMMAND_SERVICE_NAME)
	private final OzgCloudCommandService commandService;
	private final CommandMapper commandMapper;
	private final ClientAttributeService bescheidClientAttributeService;
	private final BuildProperties buildProperties;
	private final Optional<BescheidRemoteService> remoteService;

	@PostConstruct
	void logStatus() {
		remoteService.ifPresentOrElse(
				service -> LOG.info("Bescheid-Manager is configured."),
				() -> LOG.info("No BescheidRemoteService configured - Bescheid creation is not possible."));
	}

	public Bescheid createBescheid(BescheidRequest request) {
		checkRemoteService();

		return doCreateBescheid(request);
	}

	private Bescheid doCreateBescheid(BescheidRequest request) {
		var vorgang = vorgangService.getById(request.getVorgangId());
		return remoteService.map(service -> service.create(request, vorgang))
				.map(bescheid -> updateBescheid(bescheid, vorgang))
				.orElseThrow(() -> new TechnicalException(ERROR_MESSAGE_NO_SERVICE));
	}

	private Bescheid updateBescheid(Bescheid bescheid, Vorgang vorgang) {
		return bescheid.toBuilder().vorgangId(vorgang.getId()).serviceKonto(vorgang.getServiceKonto()).build();
	}

	private void checkRemoteService() {
		if (remoteService.isEmpty()) {
			LOG.error(ERROR_MESSAGE_NO_SERVICE);
			throw new TechnicalException(ERROR_MESSAGE_NO_SERVICE);
		}
	}

	public void sendBescheidManually(AttachedItem bescheidItem, Command parentCommand) {
		validateBescheidSendManually(bescheidItem, parentCommand.getRelationVersion());
		var vorgang = vorgangService.getById(VorgangId.from(bescheidItem.getVorgangId()));
		addSubCommands(parentCommand.getId(), buildSetBescheidSentSubCommands(bescheidItem, vorgang, getUserId()));
	}

	void validateBescheidSendManually(AttachedItem bescheidItem, long version) {
		validateBescheid(bescheidItem, version);
		var sendBy = bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY);
		if (Bescheid.SendBy.MANUAL.notValue(sendBy)) {
			throw new TechnicalException("Bescheid has unexpected sendBy value: '%s'. Expected is %s".formatted(sendBy, Bescheid.SendBy.MANUAL));
		}
	}

	public void sendBescheidPostfachMail(AttachedItem bescheidItem, Command parentCommand) {
		validateBescheidSendPostfach(bescheidItem, parentCommand.getRelationVersion());
		var vorgang = vorgangService.getById(VorgangId.from(bescheidItem.getVorgangId()));
		var userId = getUserId();
		var subCommands = buildSetBescheidSentSubCommands(bescheidItem, vorgang, userId);
		subCommands.add(buildSendPostfachNachrichtCommand(bescheidItem, vorgang.getServiceKonto(), userId));
		addSubCommands(parentCommand.getId(), subCommands);
	}

	public void validateBescheidSendPostfach(AttachedItem bescheidItem, long version) {
		validateBescheid(bescheidItem, version);
		var sendBy = bescheidItem.getItem().get(Bescheid.FIELD_SEND_BY);
		if (Bescheid.SendBy.NACHRICHT.notValue(sendBy)) {
			throw new TechnicalException("Bescheid has unexpected sendBy value: '%s'. Expected is %s".formatted(sendBy, Bescheid.SendBy.NACHRICHT));
		}
		if (getNachrichtSubject(bescheidItem).isEmpty()) {
			throw new TechnicalException("Bescheid has no nachricht subject");
		}
		if (getNachrichtText(bescheidItem).isEmpty()) {
			throw new TechnicalException("Bescheid has no nachricht text");
		}
	}

	void validateBescheid(AttachedItem bescheidItem, long version) {
		if (bescheidItem.getVersion() != version) {
			throw new TechnicalException("Bescheid has different version. Expected: %d, but was: %d".formatted(version, bescheidItem.getVersion()));
		}
		var status = MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_STATUS);
		if (Bescheid.Status.DRAFT.not(status)) {
			throw new TechnicalException("Bescheid has status '%s'. Bescheid must have status DRAFT".formatted(status));
		}
		if (StringUtils.isBlank(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_BESCHEID_DOCUMENT))) {
			throw new TechnicalException("Bescheid has no document");
		}
	}

	Optional<String> getNachrichtSubject(AttachedItem bescheidItem) {
		return Optional.ofNullable(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_NACHRICHT_SUBJECT)).map(StringUtils::trimToNull);
	}

	Optional<String> getNachrichtText(AttachedItem bescheidItem) {
		return Optional.ofNullable(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_NACHRICHT_TEXT)).map(StringUtils::trimToNull);
	}

	String getUserId() {
		return userProfileService.getUserProfile().getId().toString();
	}

	List<OzgCloudCommand> buildSetBescheidSentSubCommands(AttachedItem bescheidItem, Vorgang vorgang, String userId) {
		var subCommands = new ArrayList<OzgCloudCommand>();
		subCommands.add(buildVorgangBescheidenCommand(vorgang));
		subCommands.add(buildSetBescheidSentStatusCommand(bescheidItem, userId));
		return subCommands;
	}

	OzgCloudCommand buildVorgangBescheidenCommand(Vorgang vorgang) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(vorgang.getId().toString()))
				.relationId(commandMapper.mapRelationId(vorgang.getId().toString()))
				.relationVersion(vorgang.getVersion())
				.order(VORGANG_BESCHEIDEN_ORDER)
				.build();
	}

	OzgCloudCommand buildSetBescheidSentStatusCommand(AttachedItem bescheidItem, String userId) {
		return OzgCloudCommand.builder()
				.relationId(commandMapper.mapRelationId(bescheidItem.getId()))
				.relationVersion(bescheidItem.getVersion())
				.order(AttachedItemService.PATCH_ATTACHED_ITEM)
				.bodyObject(Map.of(
						AttachedItem.PROPERTY_ID, bescheidItem.getId(),
						AttachedItem.PROPERTY_ITEM, buildBescheidSentStatusItem(userId)))
				.build();
	}

	Map<String, Object> buildBescheidSentStatusItem(String userId) {
		return Map.of(
				Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name(),
				Bescheid.FIELD_SENT_INFO, buildSentInfoMap(userId));
	}

	Map<String, Object> buildSentInfoMap(String userId) {
		return Map.of(
				Bescheid.FIELD_SENT_AT, ZonedDateTime.now().format(DateTimeFormatter.ISO_DATE_TIME),
				Bescheid.FIELD_SENT_BY, userId);
	}

	OzgCloudCommand buildSendPostfachNachrichtCommand(AttachedItem bescheidItem, Vorgang.ServiceKonto serviceKonto, String userId) {
		var bodyObject = buildSendNachrichtCommandBody(bescheidItem, buildPostfachAddress(serviceKonto.getPostfachAddresses().getFirst(),
				serviceKonto.getType()));
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(bescheidItem.getVorgangId()))
				.relationId(commandMapper.mapRelationId(bescheidItem.getVorgangId()))
				.order(SEND_POSTFACH_NACHRICHT_ORDER)
				.createdBy(commandMapper.toOzgCloudUserId(userId))
				.bodyObject(bodyObject)
				.build();
	}

	Map<String, Object> buildPostfachAddress(Vorgang.PostfachAddress postfachAddress, String serviceKontoType) {
		return Map.of(Vorgang.PostfachAddress.FIELD_TYPE, postfachAddress.getType(),
				Vorgang.PostfachAddress.FIELD_VERSION, postfachAddress.getVersion(),
				Vorgang.PostfachAddress.FIELD_IDENTIFIER, postfachAddress.getIdentifier(),
				Vorgang.ServiceKonto.FIELD_SERVICEKONTO_TYPE, serviceKontoType);
	}

	Map<String, Object> buildSendNachrichtCommandBody(AttachedItem bescheidItem, Map<String, Object> postfachAddress) {
		return Map.of(
				FIELD_REPLY_OPTION, REPLY_OPTION,
				FIELD_SUBJECT, getNachrichtSubject(bescheidItem).orElse(SUBJECT),
				FIELD_MAIL_BODY, getNachrichtText(bescheidItem).orElse(StringUtils.EMPTY),
				FIELD_ATTACHMENTS, buildAttachments(bescheidItem),
				Vorgang.ServiceKonto.FIELD_POSTFACH_ADDRESS, postfachAddress);

	}

	List<String> buildAttachments(AttachedItem bescheidItem) {
		return Stream.concat(Stream.of(getBescheidFileId(bescheidItem)), getAttachments(bescheidItem).stream()).map(FileId::toString).toList();
	}

	FileId getBescheidFileId(AttachedItem bescheidItem) {
		var bescheidDocument = attachedItemService.getItem(MapUtils.getString(bescheidItem.getItem(), Bescheid.FIELD_BESCHEID_DOCUMENT));
		return FileId.from(MapUtils.getString(bescheidDocument.getItem(), Document.FIELD_DOCUMENT_FILE));
	}

	List<FileId> getAttachments(AttachedItem bescheidItem) {
		var attachmentsObject = bescheidItem.getItem().get(Bescheid.FIELD_ATTACHMENTS);
		if (attachmentsObject instanceof Collection<?> attachments) {
			return attachments.stream().map(String::valueOf).map(FileId::from).toList();
		}
		return getStringAsList(attachmentsObject);
	}

	List<FileId> getStringAsList(Object attachmentsObject) {
		if (attachmentsObject instanceof String attachment && StringUtils.isNotBlank(attachment)) {
			return Collections.singletonList(FileId.from(attachment));
		}
		return Collections.emptyList();
	}

	void addSubCommands(String parentCommandId, List<OzgCloudCommand> subCommands) {
		commandService.addSubCommands(buildCreateSubCommandsRequest(parentCommandId, subCommands));
	}

	OzgCloudCreateSubCommandsRequest buildCreateSubCommandsRequest(String parentCommandId, List<OzgCloudCommand> subCommands) {
		return OzgCloudCreateSubCommandsRequest.builder()
				.parentId(parentCommandId)
				.executionMode(SUBCOMMANDS_EXECUTION_MODE)
				.completedIfSubsCompleted(true)
				.subCommands(subCommands)
				.build();
	}

	public void setAntragBewilligung(AttachedItem bescheidItem) {
		try {
			bescheidClientAttributeService.setAntragResult(bescheidItem.getVorgangId(), getBewilligt(bescheidItem));
		} catch (Exception e) {
			LOG.error("Cannot set antrag result on vorgang (id: {})", bescheidItem.getVorgangId(), e);
		}
	}

	boolean getBewilligt(AttachedItem bescheidItem) {
		return MapUtils.getBooleanValue(bescheidItem.getItem(), Bescheid.FIELD_BEWILLIGT, false);
	}

	public BescheidManagerConfig getConfig() {
		return BescheidManagerConfig.builder()
				.version(buildProperties.getVersion())
				.javaVersion(getJavaVersion())
				.features(buildFeatures())
				.build();
	}

	String getJavaVersion() {
		return System.getProperty("java.version");
	}

	BescheidManagerConfig.Features buildFeatures() {
		return BescheidManagerConfig.Features.builder()
				.canCreateBescheidDocument(remoteService.isPresent())
				.build();
	}
}
