/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class FileServiceTest {

	@Spy
	@InjectMocks
	private FileService service;

	@Mock
	private EingangFilesRepository repository;
	@Mock
	private BinaryFileRepository binaryFileRepository;
	@Mock
	private FileIdMapper fileIdMapper;

	private final UploadedFilesReference ref = UploadedFilesReferenceTestFactory.create();
	private final Optional<String> user = Optional.empty();
	private final OzgFile file = OzgFileTestFactory.create();
	private final InputStream contentStream = new ByteArrayInputStream(IncomingFileTestFactory.CONTENT);

	@DisplayName("Upload filestream")
	@Nested
	class TestUploadFileStream {

		@Test
		void shouldCallUploadFile() {
			service.uploadFileStream(ref, file, user, contentStream);

			verify(service).uploadFile(ref, file, user, contentStream);
		}
	}

	@DisplayName("upload filestream async")
	@Nested
	class TestUploadFileStreamAsync {

		@Test
		void shouldCallUploadFile() {
			service.uploadFileStream(ref, file, user, contentStream);

			verify(service).uploadFile(ref, file, user, contentStream);
		}
	}

	@DisplayName("Upload file")
	@Nested
	class TestUploadFile {

		@BeforeEach
		void init() {
			when(binaryFileRepository.addContentStream(any(), any(), any(), any())).thenReturn(IncomingFileTestFactory.ID);
		}

		@Test
		void shouldCallAddContentSream() {
			service.uploadFileStream(ref, file, user, contentStream);

			verify(binaryFileRepository).addContentStream(ref, file, user, contentStream);
		}

		@Test
		void shouldAddContent() {
			service.uploadFileStream(ref, file, user, contentStream);

			verify(binaryFileRepository).addContentStream(ref, file, user, contentStream);
		}

		@Test
		void shouldReturnId() throws Exception {
			var id = service.uploadFileStream(ref, file, user, contentStream);

			assertThat(id).isEqualTo(IncomingFileTestFactory.ID);
		}
	}

	@Nested
	class TestGetUploadedFileContent {

		@Test
		void shouldBinaryFileRepository() {
			service.getUploadedFileStream(OzgFileTestFactory.ID);

			verify(binaryFileRepository).getFileContent(OzgFileTestFactory.ID);
		}
	}

	@Nested
	class TestFindFilesMetaData {

		private Collection<FileId> ids = Collections.singleton(OzgFileTestFactory.ID);

		@BeforeEach
		void mockFileMapper() {
			when(fileIdMapper.toString(any())).thenReturn(OzgFileTestFactory.ID.toString());
		}

		@Test
		void shouldCallFileIdMapper() {
			service.findFilesMetaData(ids);

			verify(fileIdMapper).toString(OzgFileTestFactory.ID);
		}

		@Test
		void shouldCallRepository() {
			service.findFilesMetaData(ids);

			verify(binaryFileRepository).getAll(Collections.singletonList(OzgFileTestFactory.ID.toString()));
		}
	}

	@Nested
	class TestDeleteAllByVorgang {
		@Test
		void shouldCallRepository() {
			service.deleteAllByVorgang(VorgangTestFactory.ID);

			verify(binaryFileRepository).deleteAllByVorgang(VorgangTestFactory.ID);
		}
	}
}